"use strict";
/*
 * Copyright 2020 NEM
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.ComposeService = void 0;
const LoggerFactory_1 = require("../logger/LoggerFactory");
const LogType_1 = require("../logger/LogType");
const BootstrapUtils_1 = require("./BootstrapUtils");
const _ = require("lodash");
const path_1 = require("path");
// eslint-disable-next-line @typescript-eslint/no-var-requires
const fs = require('fs');
const logger = LoggerFactory_1.default.getLogger(LogType_1.LogType.System);
const targetNodesFolder = BootstrapUtils_1.BootstrapUtils.targetNodesFolder;
const targetDatabasesFolder = BootstrapUtils_1.BootstrapUtils.targetDatabasesFolder;
const targetGatewaysFolder = BootstrapUtils_1.BootstrapUtils.targetGatewaysFolder;
class ComposeService {
    constructor(root, params) {
        this.root = root;
        this.params = params;
    }
    async run(passedPresetData) {
        const presetData = passedPresetData !== null && passedPresetData !== void 0 ? passedPresetData : BootstrapUtils_1.BootstrapUtils.loadExistingPresetData(this.params.target);
        const currentDir = process.cwd();
        const target = path_1.join(currentDir, this.params.target);
        const targetDocker = path_1.join(target, `docker`);
        if (this.params.reset) {
            BootstrapUtils_1.BootstrapUtils.deleteFolder(targetDocker);
        }
        const dockerFile = path_1.join(targetDocker, 'docker-compose.yml');
        if (fs.existsSync(dockerFile)) {
            logger.info(dockerFile + ' already exist. Reusing. (run -r to reset)');
            return BootstrapUtils_1.BootstrapUtils.loadYaml(dockerFile);
        }
        await BootstrapUtils_1.BootstrapUtils.mkdir(targetDocker);
        await BootstrapUtils_1.BootstrapUtils.generateConfiguration(presetData, path_1.join(this.root, 'config', 'docker'), targetDocker);
        const user = await BootstrapUtils_1.BootstrapUtils.resolveDockerUserFromParam(this.params.user);
        const vol = (hostFolder, imageFolder) => {
            return hostFolder + ':' + imageFolder;
        };
        logger.info(`creating docker-compose.yml from last used profile.`);
        const services = [];
        const resolvePorts = (internalPort, openPort) => {
            if (!openPort) {
                return [];
            }
            if (openPort === true || openPort === 'true') {
                return [`${internalPort}:${internalPort}`];
            }
            return [`${openPort}:${internalPort}`];
        };
        const resolveService = async (servicePreset, rawService) => {
            if (servicePreset.excludeDockerService) {
                return undefined;
            }
            if (false) {
                // POC about creating custom aws images.
                const serviceName = rawService.container_name;
                const volumes = rawService.volumes || [];
                const image = rawService.image;
                const repository = 'nem-repository';
                const dockerfileContent = `FROM docker.io/${image}\n\n${volumes
                    .map((v) => {
                    const parts = v.split(':');
                    return `ADD ${parts[0].replace('../', '').replace('./', 'docker/')} ${parts[1]}`;
                })
                    .join('\n')}\n`;
                const dockerFile = path_1.join(target, 'Dockerfile-' + serviceName);
                await BootstrapUtils_1.BootstrapUtils.writeTextFile(dockerFile, dockerfileContent);
                await Promise.all(volumes.map(async (v) => {
                    const parts = v.split(':');
                    await BootstrapUtils_1.BootstrapUtils.mkdir(path_1.join(targetDocker, parts[0]));
                }));
                const generatedImageName = repository + ':' + serviceName;
                await BootstrapUtils_1.BootstrapUtils.createImageUsingExec(target, dockerFile, generatedImageName);
                // const awsUserId = '172617417348';
                // aws ecr get-login-password --region us-east-1 | docker login --username AWS --password-stdin 172617417348.dkr.ecr.us-east-1.amazonaws.com
                // const absoluteImageUrl = `${awsUserId}.dkr.ecr.us-east-1.amazonaws.com/${generatedImageName}`;
                // await BootstrapUtils.exec(`docker tag ${generatedImageName} ${absoluteImageUrl}`);
                return Object.assign(Object.assign({}, rawService), { image: generatedImageName, volumes: undefined });
            }
            else {
                const service = Object.assign({}, rawService);
                if (servicePreset.host || servicePreset.ipv4_address) {
                    service.networks = { default: {} };
                }
                if (servicePreset.host) {
                    service.hostname = servicePreset.host;
                    service.networks.default.aliases = [servicePreset.host];
                }
                if (servicePreset.environment) {
                    service.environment = Object.assign(Object.assign({}, servicePreset.environment), rawService.environment);
                }
                if (servicePreset.ipv4_address) {
                    service.networks.default.ipv4_address = servicePreset.ipv4_address;
                }
                return service;
            }
        };
        await Promise.all((presetData.databases || []).map(async (n) => {
            services.push(await resolveService(n, {
                user,
                environment: { MONGO_INITDB_DATABASE: n.databaseName || presetData.databaseName },
                container_name: n.name,
                image: presetData.mongoImage,
                command: `mongod --dbpath=/dbdata --bind_ip=${n.name}`,
                stop_signal: 'SIGINT',
                working_dir: '/docker-entrypoint-initdb.d',
                ports: resolvePorts(27017, n.openPort),
                volumes: [
                    vol(`./mongo`, `/docker-entrypoint-initdb.d/:ro`),
                    vol(`../${targetDatabasesFolder}/${n.name}`, '/dbdata:rw'),
                ],
            }));
        }));
        const nodeWorkingDirectory = '/symbol-workdir';
        const nodeCommandsDirectory = '/symbol-commands';
        await Promise.all((presetData.nodes || []).map(async (n) => {
            const nodeService = await resolveService(n, {
                user,
                container_name: n.name,
                image: presetData.symbolServerImage,
                command: `bash -c "/bin/bash ${nodeCommandsDirectory}/runServerRecover.sh  ${n.name} && /bin/bash ${nodeCommandsDirectory}/startServer.sh ${n.name}"`,
                stop_signal: 'SIGINT',
                working_dir: nodeWorkingDirectory,
                restart: 'on-failure:2',
                ports: resolvePorts(7900, n.openPort),
                volumes: [vol(`../${targetNodesFolder}/${n.name}`, nodeWorkingDirectory), vol(`./server`, nodeCommandsDirectory)],
                depends_on: n.brokerName ? [n.brokerName] : [],
            });
            services.push(nodeService);
            if (n.brokerName && nodeService) {
                services.push(await resolveService({
                    ipv4_address: n.brokerIpv4_address,
                    openPort: n.brokerOpenPort,
                    excludeDockerService: n.brokerExcludeDockerService,
                    host: n.brokerHost,
                }, {
                    user,
                    container_name: n.brokerName,
                    image: nodeService.image,
                    working_dir: nodeWorkingDirectory,
                    command: `bash -c "/bin/bash ${nodeCommandsDirectory}/runServerRecover.sh ${n.brokerName} && /bin/bash ${nodeCommandsDirectory}/startBroker.sh ${n.brokerName}"`,
                    ports: resolvePorts(7902, n.brokerOpenPort),
                    stop_signal: 'SIGINT',
                    restart: 'on-failure:2',
                    volumes: nodeService.volumes,
                }));
            }
        }));
        await Promise.all((presetData.gateways || []).map(async (n) => {
            services.push(await resolveService(n, {
                container_name: n.name,
                user,
                image: presetData.symbolRestImage,
                command: 'npm start --prefix /app/catapult-rest/rest /symbol-workdir/rest.json',
                stop_signal: 'SIGINT',
                working_dir: nodeWorkingDirectory,
                ports: resolvePorts(3000, n.openPort),
                volumes: [vol(`../${targetGatewaysFolder}/${n.name}`, nodeWorkingDirectory)],
                depends_on: [n.databaseHost],
            }));
        }));
        const validServices = services.filter((s) => s).map((s) => s);
        const servicesMap = _.keyBy(validServices, 'container_name');
        const dockerCompose = {
            version: '3',
            services: servicesMap,
        };
        if (presetData.subnet)
            dockerCompose.networks = {
                default: {
                    ipam: {
                        config: [
                            {
                                subnet: presetData.subnet,
                            },
                        ],
                    },
                },
            };
        await BootstrapUtils_1.BootstrapUtils.writeYaml(dockerFile, dockerCompose);
        logger.info(`docker-compose.yml file created ${dockerFile}`);
        return dockerCompose;
    }
}
exports.ComposeService = ComposeService;
ComposeService.defaultParams = {
    target: BootstrapUtils_1.BootstrapUtils.defaultTargetFolder,
    user: BootstrapUtils_1.BootstrapUtils.CURRENT_USER,
    reset: false,
};
