"use strict";
/*
 * Copyright 2020 NEM
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.ConfigService = exports.Preset = void 0;
const BootstrapUtils_1 = require("./BootstrapUtils");
const symbol_sdk_1 = require("symbol-sdk");
const CertificateService_1 = require("./CertificateService");
const LoggerFactory_1 = require("../logger/LoggerFactory");
const LogType_1 = require("../logger/LogType");
const NemgenService_1 = require("./NemgenService");
const model_1 = require("../model");
const fs = require("fs");
const VotingService_1 = require("./VotingService");
const path_1 = require("path");
const ReportService_1 = require("./ReportService");
/**
 * Defined presets.
 */
var Preset;
(function (Preset) {
    Preset["bootstrap"] = "bootstrap";
    Preset["testnet"] = "testnet";
})(Preset = exports.Preset || (exports.Preset = {}));
const logger = LoggerFactory_1.default.getLogger(LogType_1.LogType.System);
class ConfigService {
    constructor(root, params) {
        this.root = root;
        this.params = params;
    }
    toConfig(account) {
        return {
            privateKey: account.privateKey,
            publicKey: account.publicAccount.publicKey,
            address: account.address.plain(),
        };
    }
    static getNetworkIdentifier(networkType) {
        switch (networkType) {
            case symbol_sdk_1.NetworkType.MAIN_NET:
                return 'public';
            case symbol_sdk_1.NetworkType.TEST_NET:
                return 'public-test';
            case symbol_sdk_1.NetworkType.MIJIN:
                return 'mijin';
            case symbol_sdk_1.NetworkType.MIJIN_TEST:
                return 'mijin-test';
        }
        throw new Error(`Invalid Network Type ${networkType}`);
    }
    static getNetworkName(networkType) {
        switch (networkType) {
            case symbol_sdk_1.NetworkType.MAIN_NET:
                return 'public';
            case symbol_sdk_1.NetworkType.TEST_NET:
                return 'publicTest';
            case symbol_sdk_1.NetworkType.MIJIN:
                return 'mijin';
            case symbol_sdk_1.NetworkType.MIJIN_TEST:
                return 'mijinTest';
        }
        throw new Error(`Invalid Network Type ${networkType}`);
    }
    generateAddresses(networkType, size) {
        return ConfigService.getArray(size).map(() => this.toConfig(symbol_sdk_1.Account.generateNewAccount(networkType)));
    }
    generateAccount(networkType, privateKey) {
        return privateKey ? symbol_sdk_1.Account.createFromPrivateKey(privateKey, networkType) : symbol_sdk_1.Account.generateNewAccount(networkType);
    }
    async generateNodeAccount(index, nodePreset, networkType) {
        const name = nodePreset.name || `node-${index}`;
        const { ssl, node } = await new CertificateService_1.CertificateService(this.root, this.params).run(name);
        const friendlyName = nodePreset.friendlyName || ssl.publicKey.substr(0, 7);
        const nodeAccount = { name, friendlyName, roles: nodePreset.roles, ssl, node };
        if (nodePreset.harvesting || nodePreset.voting)
            nodeAccount.signing = this.toConfig(this.generateAccount(networkType, nodePreset.signingPrivateKey));
        if (nodePreset.voting)
            nodeAccount.voting = this.toConfig(this.generateAccount(networkType, nodePreset.votingPrivateKey));
        if (nodePreset.harvesting)
            nodeAccount.vrf = this.toConfig(this.generateAccount(networkType, nodePreset.vrfPrivateKey));
        return nodeAccount;
    }
    async generateNodeAccounts(networkType, nodes) {
        return Promise.all(nodes.map((node, index) => this.generateNodeAccount(index, node, networkType)));
    }
    static getArray(size) {
        return [...Array(size).keys()];
    }
    async run() {
        const target = this.params.target;
        try {
            if (this.params.reset) {
                BootstrapUtils_1.BootstrapUtils.deleteFolder(target);
            }
            const presetLocation = BootstrapUtils_1.BootstrapUtils.getGeneratedPresetLocation(target);
            if (fs.existsSync(presetLocation)) {
                logger.info(`The generated preset ${presetLocation} already exist, ignoring configuration. (run -r to reset)`);
                const presetData = BootstrapUtils_1.BootstrapUtils.loadExistingPresetData(target);
                const addresses = BootstrapUtils_1.BootstrapUtils.loadExistingAddresses(target);
                return { presetData, addresses };
            }
            const presetData = BootstrapUtils_1.BootstrapUtils.loadPresetData(this.root, this.params.preset, this.params.assembly, this.params.customPreset, this.params.customPresetObject);
            await BootstrapUtils_1.BootstrapUtils.pullImage(presetData.symbolServerToolsImage);
            const networkType = presetData.networkType;
            const addresses = await this.generateRandomConfiguration(networkType, presetData);
            await BootstrapUtils_1.BootstrapUtils.writeYaml(BootstrapUtils_1.BootstrapUtils.getGeneratedAddressLocation(target), addresses);
            await this.generateNodes(presetData, addresses);
            await this.generateNemesis(presetData, addresses);
            await this.generateGateways(presetData, addresses);
            await BootstrapUtils_1.BootstrapUtils.writeYaml(presetLocation, presetData);
            logger.info(`Configuration generated.`);
            return { presetData, addresses };
        }
        catch (e) {
            logger.error(`Unknown error generating the configuration. ${e.message}`, e);
            logger.error(`The target folder '${target}' should be deleted!!!`);
            throw e;
        }
    }
    async generateNemesis(presetData, addresses) {
        const target = this.params.target;
        const nemesisSeedFolder = BootstrapUtils_1.BootstrapUtils.getTargetNemesisFolder(target, false, 'seed');
        if (presetData.nemesis) {
            await this.generateNemesisConfig(presetData, addresses);
        }
        else {
            const copyFrom = presetData.nemesisSeedFolder || path_1.join(this.root, 'presets', this.params.preset, 'seed');
            await BootstrapUtils_1.BootstrapUtils.generateConfiguration({}, copyFrom, nemesisSeedFolder);
        }
        if (this.params.report) {
            await new ReportService_1.ReportService(this.root, this.params).run(presetData);
        }
        BootstrapUtils_1.BootstrapUtils.validateFolder(nemesisSeedFolder);
        await Promise.all((addresses.nodes || []).map(async (account) => {
            const name = account.name;
            const dataFolder = BootstrapUtils_1.BootstrapUtils.getTargetNodesFolder(target, false, name, 'data');
            await BootstrapUtils_1.BootstrapUtils.generateConfiguration({}, nemesisSeedFolder, dataFolder);
        }));
        await BootstrapUtils_1.BootstrapUtils.writeYaml(BootstrapUtils_1.BootstrapUtils.getGeneratedPresetLocation(target), presetData);
        return { presetData, addresses };
    }
    async generateRandomConfiguration(networkType, presetData) {
        const addresses = {
            networkType: networkType,
            nemesisGenerationHashSeed: presetData.nemesisGenerationHashSeed || symbol_sdk_1.Account.generateNewAccount(networkType).publicAccount.publicKey,
        };
        if (presetData.nodes) {
            addresses.nodes = await this.generateNodeAccounts(networkType, presetData.nodes);
        }
        if (presetData.gateways) {
            addresses.gateways = this.generateAddresses(networkType, presetData.gateways.length);
        }
        const sinkAddress = symbol_sdk_1.Account.generateNewAccount(networkType).address.plain();
        if (!presetData.harvestNetworkFeeSinkAddress) {
            presetData.harvestNetworkFeeSinkAddress = sinkAddress;
        }
        if (!presetData.mosaicRentalFeeSinkAddress) {
            presetData.mosaicRentalFeeSinkAddress = sinkAddress;
        }
        if (!presetData.namespaceRentalFeeSinkAddress) {
            presetData.namespaceRentalFeeSinkAddress = sinkAddress;
        }
        presetData.networkIdentifier = ConfigService.getNetworkIdentifier(presetData.networkType);
        presetData.networkName = ConfigService.getNetworkName(presetData.networkType);
        if (!presetData.nemesisGenerationHashSeed) {
            presetData.nemesisGenerationHashSeed = addresses.nemesisGenerationHashSeed;
        }
        if (presetData.nemesis) {
            addresses.nemesisSigner = this.toConfig(this.generateAccount(networkType, presetData.nemesis.nemesisSignerPrivateKey));
            if (!presetData.nemesis.nemesisSignerPrivateKey && addresses.nemesisSigner) {
                presetData.nemesis.nemesisSignerPrivateKey = addresses.nemesisSigner.privateKey;
            }
        }
        if (!presetData.nemesisSignerPublicKey && addresses.nemesisSigner) {
            presetData.nemesisSignerPublicKey = addresses.nemesisSigner.publicKey;
        }
        const nemesisSignerAddress = symbol_sdk_1.Address.createFromPublicKey(presetData.nemesisSignerPublicKey, networkType);
        if (!presetData.currencyMosaicId)
            presetData.currencyMosaicId = BootstrapUtils_1.BootstrapUtils.toHex(symbol_sdk_1.MosaicId.createFromNonce(symbol_sdk_1.MosaicNonce.createFromNumber(0), nemesisSignerAddress).toHex());
        if (!presetData.harvestingMosaicId) {
            if (!presetData.nemesis) {
                throw new Error('nemesis must be defined!');
            }
            if (presetData.nemesis.mosaics && presetData.nemesis.mosaics.length > 1) {
                presetData.harvestingMosaicId = BootstrapUtils_1.BootstrapUtils.toHex(symbol_sdk_1.MosaicId.createFromNonce(symbol_sdk_1.MosaicNonce.createFromNumber(1), nemesisSignerAddress).toHex());
            }
            else {
                presetData.harvestingMosaicId = presetData.currencyMosaicId;
            }
        }
        if (presetData.nemesis) {
            if (presetData.nemesis.mosaics) {
                const mosaics = [];
                presetData.nemesis.mosaics.forEach((m, index) => {
                    const accounts = this.generateAddresses(networkType, m.accounts);
                    mosaics.push({
                        id: index ? presetData.currencyMosaicId : presetData.harvestingMosaicId,
                        name: m.name,
                        type: index ? 'harvest' : 'currency',
                        accounts,
                    });
                });
                presetData.nemesis.mosaics.forEach((m, index) => {
                    const accounts = mosaics[index].accounts;
                    if (!m.currencyDistributions) {
                        const signingNodes = (addresses.nodes || []).filter((node) => node.signing);
                        const totalAccounts = (m.accounts || 0) + signingNodes.length;
                        const amountPerAccount = Math.floor(m.supply / totalAccounts);
                        m.currencyDistributions = [
                            ...accounts.map((a) => ({ address: a.address, amount: amountPerAccount })),
                            ...signingNodes.map((n) => ({ address: n.signing.address, amount: amountPerAccount })),
                        ];
                        if (m.currencyDistributions.length)
                            m.currencyDistributions[0].amount += m.supply - totalAccounts * amountPerAccount;
                    }
                    const supplied = m.currencyDistributions.map((d) => d.amount).reduce((a, b) => a + b, 0);
                    if (m.supply != supplied) {
                        throw new Error(`Invalid nemgen total supplied value, expected ${m.supply} but total is ${supplied}`);
                    }
                });
                addresses.mosaics = mosaics;
            }
        }
        return addresses;
    }
    async generateNodes(presetData, addresses) {
        await Promise.all((addresses.nodes || []).map(async (account, index) => await this.generateNodeConfiguration(account, index, presetData, addresses)));
    }
    async generateNodeConfiguration(account, index, presetData, addresses) {
        var _a, _b;
        const copyFrom = path_1.join(this.root, 'config', 'node');
        const name = account.name;
        const outputFolder = BootstrapUtils_1.BootstrapUtils.getTargetNodesFolder(this.params.target, false, name, 'userconfig');
        const nodePreset = (presetData.nodes || [])[index];
        const generatedContext = {
            name: name,
            friendlyName: (nodePreset === null || nodePreset === void 0 ? void 0 : nodePreset.friendlyName) || account.friendlyName,
            harvesterSigningPrivateKey: ((_a = account.signing) === null || _a === void 0 ? void 0 : _a.privateKey) || '',
            harvesterVrfPrivateKey: ((_b = account.vrf) === null || _b === void 0 ? void 0 : _b.privateKey) || '',
        };
        const templateContext = Object.assign(Object.assign(Object.assign({}, presetData), generatedContext), nodePreset);
        await BootstrapUtils_1.BootstrapUtils.generateConfiguration(templateContext, copyFrom, outputFolder);
        await this.generateP2PFile(presetData, addresses, outputFolder, model_1.NodeType.PEER_NODE, (nodePresetData) => nodePresetData.harvesting, 'peers-p2p.json');
        await this.generateP2PFile(presetData, addresses, outputFolder, model_1.NodeType.API_NODE, (nodePresetData) => nodePresetData.api, 'peers-api.json');
        await new VotingService_1.VotingService(this.params).run(presetData, account, nodePreset);
    }
    async generateP2PFile(presetData, addresses, outputFolder, type, nodePresetDataFunction, jsonFileName) {
        var _a;
        const thisNetworkKnownPeers = (presetData.nodes || [])
            .map((nodePresetData, index) => {
            if (!nodePresetDataFunction(nodePresetData)) {
                return undefined;
            }
            const node = (addresses.nodes || [])[index];
            return {
                publicKey: node.ssl.publicKey,
                endpoint: {
                    host: nodePresetData.host,
                    port: 7900,
                },
                metadata: {
                    name: nodePresetData.friendlyName,
                    roles: nodePresetData.roles,
                },
            };
        })
            .filter((i) => i);
        const globalKnownPeers = ((_a = presetData.knownPeers) === null || _a === void 0 ? void 0 : _a[type]) || [];
        const data = {
            _info: `this file contains a list of ${type} peers`,
            knownPeers: [...thisNetworkKnownPeers, ...globalKnownPeers],
        };
        await fs.promises.writeFile(path_1.join(outputFolder, `resources`, jsonFileName), JSON.stringify(data, null, 2));
    }
    async generateNemesisConfig(presetData, addresses) {
        if (!presetData.nemesis) {
            throw new Error('nemesis must not be defined!');
        }
        const target = this.params.target;
        const nemesisWorkingDir = BootstrapUtils_1.BootstrapUtils.getTargetNemesisFolder(target, false);
        const transactionsDirectory = path_1.join(nemesisWorkingDir, presetData.nemesis.transactionsDirectory || presetData.transactionsDirectory);
        await BootstrapUtils_1.BootstrapUtils.mkdir(transactionsDirectory);
        const copyFrom = path_1.join(this.root, `config`, `nemesis`);
        const moveTo = path_1.join(nemesisWorkingDir, `userconfig`);
        const templateContext = Object.assign(Object.assign({}, presetData), { addresses });
        await Promise.all((addresses.nodes || []).filter((n) => n.vrf).map((n) => this.createVrfTransaction(transactionsDirectory, presetData, n)));
        await Promise.all((addresses.nodes || [])
            .filter((n) => n.voting)
            .map((n) => this.createVotingKeyTransaction(transactionsDirectory, presetData, n)));
        if (presetData.nemesis.mosaics && (presetData.nemesis.transactions || presetData.nemesis.balances)) {
            logger.info('Opt In mode is ON!!! balances or transactions have been provided');
            if (presetData.nemesis.transactions) {
                const transactionHashes = [];
                const transactions = (await Promise.all(Object.entries(presetData.nemesis.transactions || {})
                    .map(([key, payload]) => {
                    const transactionHash = symbol_sdk_1.Transaction.createTransactionHash(payload, Array.from(symbol_sdk_1.Convert.hexToUint8(presetData.nemesisGenerationHashSeed)));
                    if (transactionHashes.indexOf(transactionHash) > -1) {
                        logger.warn(`Transaction ${key} wth hash ${transactionHash} already exist. Excluded from folder.`);
                        return undefined;
                    }
                    transactionHashes.push(transactionHash);
                    return this.storeTransaction(transactionsDirectory, key, payload);
                })
                    .filter((p) => p))).filter((p) => p);
                logger.info(`Found ${transactions.length} opted in transactions.`);
            }
            const currencyMosaic = presetData.nemesis.mosaics[0];
            const nglAccount = currencyMosaic.currencyDistributions[0];
            const originalNglAccountBalance = nglAccount.amount;
            if (!nglAccount) {
                throw Error('"NGL" account could not be found for opt in!');
            }
            let totalOptedInBalance = 0;
            if (presetData.nemesis.balances) {
                Object.entries(presetData.nemesis.balances || {}).forEach(([address, amount]) => {
                    totalOptedInBalance += amount;
                    currencyMosaic.currencyDistributions.push({ address, amount });
                });
                logger.info(`Removing ${Object.keys(presetData.nemesis.balances).length} accounts (total of ${totalOptedInBalance}) from "ngl" account ${nglAccount.address}`);
            }
            nglAccount.amount = nglAccount.amount - totalOptedInBalance;
            const providedBalances = Object.values(currencyMosaic.currencyDistributions)
                .map((d) => d.amount)
                .reduce((a, b) => a + b, 0);
            const currentBalance = providedBalances;
            if (nglAccount.amount < 1) {
                throw new Error(`NGL account didn't have enough balance (${originalNglAccountBalance}) to paid all the supplied optedin namespaces and accounts of ${currentBalance}`);
            }
            if (currentBalance !== currencyMosaic.supply) {
                throw new Error(`Current supplied balance of ${currentBalance} is different from expected supply of ${currencyMosaic.supply}`);
            }
        }
        await BootstrapUtils_1.BootstrapUtils.generateConfiguration(templateContext, copyFrom, moveTo);
        await new NemgenService_1.NemgenService(this.root, this.params).run(presetData);
    }
    async createVrfTransaction(transactionsDirectory, presetData, node) {
        if (!node.vrf) {
            throw new Error('VRF keys should have been generated!!');
        }
        if (!node.signing) {
            throw new Error('Signing keys should have been generated!!');
        }
        const deadline = symbol_sdk_1.Deadline.createFromDTO('1');
        const vrf = symbol_sdk_1.VrfKeyLinkTransaction.create(deadline, node.vrf.publicKey, symbol_sdk_1.LinkAction.Link, presetData.networkType, symbol_sdk_1.UInt64.fromUint(0));
        const account = symbol_sdk_1.Account.createFromPrivateKey(node.signing.privateKey, presetData.networkType);
        const signedTransaction = account.sign(vrf, presetData.nemesisGenerationHashSeed);
        return await this.storeTransaction(transactionsDirectory, `vrf_${node.name}`, signedTransaction.payload);
    }
    async createVotingKeyTransaction(transactionsDirectory, presetData, node) {
        if (!node.voting) {
            throw new Error('Voting keys should have been generated!!');
        }
        if (!node.signing) {
            throw new Error('Signing keys should have been generated!!');
        }
        const deadline = symbol_sdk_1.Deadline.createFromDTO('1');
        const votingKey = BootstrapUtils_1.BootstrapUtils.createVotingKey(node.voting.publicKey);
        const voting = symbol_sdk_1.VotingKeyLinkTransaction.create(deadline, votingKey, presetData.votingKeyStartEpoch, presetData.votingKeyEndEpoch, symbol_sdk_1.LinkAction.Link, presetData.networkType, symbol_sdk_1.UInt64.fromUint(0));
        const account = symbol_sdk_1.Account.createFromPrivateKey(node.signing.privateKey, presetData.networkType);
        const signedTransaction = account.sign(voting, presetData.nemesisGenerationHashSeed);
        return await this.storeTransaction(transactionsDirectory, `voting_${node.name}`, signedTransaction.payload);
    }
    async storeTransaction(transactionsDirectory, name, payload) {
        const transaction = symbol_sdk_1.TransactionMapping.createFromPayload(payload);
        await fs.promises.writeFile(`${transactionsDirectory}/${name}.bin`, symbol_sdk_1.Convert.hexToUint8(payload));
        return transaction;
    }
    generateGateways(presetData, addresses) {
        return Promise.all((addresses.gateways || []).map(async (account, index) => {
            const copyFrom = path_1.join(this.root, 'config', 'rest-gateway');
            const generatedContext = {
                restPrivateKey: account.privateKey,
            };
            const gatewayPreset = (presetData.gateways || [])[index];
            const templateContext = Object.assign(Object.assign(Object.assign({}, presetData), generatedContext), gatewayPreset);
            const name = templateContext.name || `rest-gateway-${index}`;
            const moveTo = BootstrapUtils_1.BootstrapUtils.getTargetGatewayFolder(this.params.target, false, name);
            await BootstrapUtils_1.BootstrapUtils.generateConfiguration(templateContext, copyFrom, moveTo);
            const apiNodeConfigFolder = BootstrapUtils_1.BootstrapUtils.getTargetNodesFolder(this.params.target, false, gatewayPreset.apiNodeName, 'userconfig', 'resources');
            await BootstrapUtils_1.BootstrapUtils.generateConfiguration({}, apiNodeConfigFolder, path_1.join(moveTo, 'api-node-config'));
        }));
    }
}
exports.ConfigService = ConfigService;
ConfigService.defaultParams = {
    target: BootstrapUtils_1.BootstrapUtils.defaultTargetFolder,
    report: false,
    preset: Preset.bootstrap,
    reset: false,
    user: BootstrapUtils_1.BootstrapUtils.CURRENT_USER,
};
