"use strict";
/*
 * Copyright 2020 NEM
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.LinkService = void 0;
const LoggerFactory_1 = require("../logger/LoggerFactory");
const LogType_1 = require("../logger/LogType");
const symbol_sdk_1 = require("symbol-sdk");
const BootstrapUtils_1 = require("./BootstrapUtils");
const _ = require("lodash");
const operators_1 = require("rxjs/operators");
const fromArray_1 = require("rxjs/internal/observable/fromArray");
const rxjs_1 = require("rxjs");
const logger = LoggerFactory_1.default.getLogger(LogType_1.LogType.System);
class LinkService {
    constructor(params) {
        this.params = params;
    }
    async run(passedPresetData, passedAddresses) {
        const presetData = passedPresetData !== null && passedPresetData !== void 0 ? passedPresetData : BootstrapUtils_1.BootstrapUtils.loadExistingPresetData(this.params.target);
        const addresses = passedAddresses !== null && passedAddresses !== void 0 ? passedAddresses : BootstrapUtils_1.BootstrapUtils.loadExistingAddresses(this.params.target);
        const url = this.params.url.replace(/\/$/, '');
        logger.info(`${this.params.unlink ? 'Unlinking' : 'Linking'} nodes using network url ${url}. Max Fee ${this.params.maxFee / Math.pow(10, symbol_sdk_1.NetworkCurrencyPublic.DIVISIBILITY)}`);
        const repositoryFactory = new symbol_sdk_1.RepositoryFactoryHttp(url);
        const generationHash = await repositoryFactory.getGenerationHash().toPromise();
        if (generationHash !== presetData.nemesisGenerationHashSeed) {
            throw new Error(`You are connecting to the wrong network. Expected generation hash is ${presetData.nemesisGenerationHashSeed} but got ${generationHash}`);
        }
        const transactionNodes = this.createTransactionsToAnnounce(addresses, presetData);
        if (!transactionNodes.length) {
            logger.info(`There are no transactions to announce...`);
            return;
        }
        const transactionRepository = repositoryFactory.createTransactionRepository();
        const transactionService = new symbol_sdk_1.TransactionService(transactionRepository, repositoryFactory.createReceiptRepository());
        const listener = repositoryFactory.createListener();
        await listener.open();
        const faucetUrl = presetData.faucetUrl;
        const signedTransactionObservable = fromArray_1.fromArray(transactionNodes).pipe(operators_1.mergeMap(({ node, transactions }) => {
            if (!node.signing) {
                throw new Error('Signing is required!');
            }
            const account = symbol_sdk_1.Account.createFromPrivateKey(node.signing.privateKey, presetData.networkType);
            const noFundsMessage = faucetUrl
                ? `Does your node signing address have any network coin? Send some tokens to ${account.address.plain()} via ${faucetUrl}`
                : `Does your node signing address have any network coin? Send some tokens to ${account.address.plain()} .`;
            return repositoryFactory
                .createAccountRepository()
                .getAccountInfo(account.address)
                .pipe(operators_1.mergeMap((a) => {
                const currencyMosaicIdHex = BootstrapUtils_1.BootstrapUtils.toHex(presetData.currencyMosaicId);
                const mosaic = a.mosaics.find((m) => BootstrapUtils_1.BootstrapUtils.toHex(m.id.toHex()) === currencyMosaicIdHex);
                if (!mosaic || mosaic.amount.compare(symbol_sdk_1.UInt64.fromUint(0)) < 1) {
                    logger.error(`Node signing account ${account.address.plain()} does not have enough currency. Mosaic id: ${currencyMosaicIdHex}. \n\n${noFundsMessage}`);
                    return rxjs_1.EMPTY;
                }
                return fromArray_1.fromArray(transactions.map((t) => account.sign(t, generationHash)));
            }), operators_1.catchError((e) => {
                logger.error(`Node signing account ${account.address.plain()} is not valid. ${e.message}. \n\n${noFundsMessage}`);
                return rxjs_1.EMPTY;
            }));
        }));
        const announceCalls = signedTransactionObservable.pipe(operators_1.mergeMap((signedTransaction) => {
            return transactionService.announce(signedTransaction, listener).pipe(operators_1.map((completedTransaction) => {
                var _a, _b;
                const message = `Transaction ${completedTransaction.type} ${(_a = completedTransaction.transactionInfo) === null || _a === void 0 ? void 0 : _a.hash} - signer ${(_b = completedTransaction.signer) === null || _b === void 0 ? void 0 : _b.address.plain()} has been confirmed`;
                logger.info(message);
                return message;
            }), operators_1.catchError((e) => {
                const message = `Transaction ${signedTransaction.type} ${signedTransaction.hash} - signer ${signedTransaction.getSignerAddress().plain()} failed!! ` + e.message;
                logger.error(message);
                return rxjs_1.of(message);
            }));
        }));
        await announceCalls.pipe(operators_1.toArray()).toPromise();
        listener.close();
    }
    createTransactionsToAnnounce(addresses, presetData) {
        return _.flatMap(addresses.nodes || [])
            .filter((node) => node.signing)
            .map((node) => {
            const transactions = [];
            if (!node.signing) {
                throw new Error('Signing is required!');
            }
            const account = symbol_sdk_1.Account.createFromPrivateKey(node.signing.privateKey, presetData.networkType);
            const action = this.params.unlink ? symbol_sdk_1.LinkAction.Unlink : symbol_sdk_1.LinkAction.Link;
            if (node.vrf) {
                logger.info(`Creating VrfKeyLinkTransaction - node: ${node.name}, signer public key: ${account.publicKey}, vrf key: ${node.vrf.publicKey}`);
                transactions.push(symbol_sdk_1.VrfKeyLinkTransaction.create(symbol_sdk_1.Deadline.create(), node.vrf.publicKey, action, presetData.networkType, symbol_sdk_1.UInt64.fromUint(this.params.maxFee)));
            }
            if (node.voting) {
                const votingPublicKey = BootstrapUtils_1.BootstrapUtils.createVotingKey(node.voting.publicKey);
                logger.info(`Creating VotingKeyLinkTransaction - node: ${node.name}, signer public key: ${account.publicKey}, voting public key: ${votingPublicKey}`);
                transactions.push(symbol_sdk_1.VotingKeyLinkTransaction.create(symbol_sdk_1.Deadline.create(), votingPublicKey, presetData.votingKeyStartEpoch, presetData.votingKeyEndEpoch, action, presetData.networkType, symbol_sdk_1.UInt64.fromUint(this.params.maxFee)));
            }
            return { node, transactions };
        });
    }
}
exports.LinkService = LinkService;
LinkService.defaultParams = {
    target: BootstrapUtils_1.BootstrapUtils.defaultTargetFolder,
    url: 'http://localhost:3000',
    maxFee: 100000,
    unlink: false,
};
